/*
 * This file contains a driver for the Hard Disk Controller (HDC)
 * on the Atari ST.
 *
 * The driver supports two operations: read a block and write a block.
 * It accepts two messages, one for reading and one for writing,
 * both using message format m2 and with the same parameters:
 *
 *    m_type      DEVICE    PROC_NR     COUNT    POSITION  ADRRESS
 * ----------------------------------------------------------------
 * |  DEV_READ  | device  | proc nr |  bytes  |  offset | buf ptr |
 * |------------+---------+---------+---------+---------+---------|
 * | DEV_WRITE  | device  | proc nr |  bytes  |  offset | buf ptr |
 * |--------------------------------------------------------------|
 * |SCATTERED_IO| device  | proc nr | requests|         | iov ptr |
#if (HD_CLOCK == 1)
 * |------------+---------+---------+---------+---------+---------|
 * | DEV_IOCTL  | device  | proc nr |func code|         | address |
#endif 
 * |------------+---------+---------+---------+---------+---------|
 * | DEV_OPEN   | device  | proc nr |         |         |         |
 * |------------+---------+---------+---------+---------+---------|
 * | DEV_CLOSE  | device  | proc nr |         |         |         |
 * ----------------------------------------------------------------
 *
 * The file contains one entry point:
 *
 *	winchester_task: main entry when system is brought up
 *
 * A single ST may have several controllers.
 * Each controller support up to two drives.
 * TODO:
 * - initialisation?
 * - assure usage of ST ram
 */

#include "kernel.h"
#if (MACHINE == ATARI)
#include <minix/callnr.h>
#include <minix/com.h>
#include "proc.h"

#include "staddr.h"
#include "stmfp.h"
#include "sthdc.h"
#include "stdma.h"
#include <sgtty.h>

#if (NR_ACSI_DRIVES != 0)

#define TRACE(x)	/* x */

/* Parameters for the disk drive. */
#define SECTOR_SIZE	512	/* physical sector size in bytes */
#define MINOR_SHIFT	4	/* 2 ** MINOR_SHIFT == #minors/drive */
#define XFERSIZE 13		/* # bytes to be sent to the hd clocks */

/* dis/enable interrupts */
#define IENABLE()	MFP->mf_ierb |= IB_DINT
#define IDISABLE()	MFP->mf_ierb &= ~IB_DINT

/* timing constants */
#define	HDC_DELAY	0	/* for dma[rw]dat() */
#define	HDC_POLL	30000	/* polling of MFP for interrupt bit */

#if (FAST_DISK == 1)
# define cmd_delay()
#else
# define COMMAND_DELAY	1000	/* delay between two acsi commands */
# define cmd_delay()	if (1) { int i; for (i = 0; i < COMMAND_DELAY; i++) ; } else
#endif
PRIVATE int cmd_delay_enabled = 0;

PRIVATE message mess;		/* message buffer for in and out */

FORWARD _PROTOTYPE( int do_ioctl, (message *mp)				);
FORWARD _PROTOTYPE( void flush_dmaq, (void)				);
FORWARD _PROTOTYPE( int poll, (void)					);
FORWARD _PROTOTYPE( void hdcint, (void)					);

/*===========================================================================*
 *				winchester_task				     *
 *===========================================================================*/
PUBLIC void winchester_task()
{
  register int r, caller, procno;

  TRACE(printf("winchester: task started\n"));

  /*
   * The main loop of the disk task.
   * It waits for a message, carries it out, and sends a reply.
   */
  while (TRUE) {
	cmd_delay_enabled = 0;
	receive(ANY, &mess);
	if (mess.m_source < 0)
		panic("disk task got message from ", mess.m_source);
	TRACE(printf("hd: received %d from %d\n",mess.m_type,mess.m_source));
	caller = mess.m_source;
	procno = mess.PROC_NR;

	/* Now carry out the work. */
	switch (mess.m_type) {
	    case DEV_OPEN:	r = wini_open(&mess); break;
	    case DEV_CLOSE:	r = wini_close(&mess);	break;
	    case DEV_READ:
	    case DEV_WRITE:	r = wini_rdwt(&mess);	break;
#if (ACSI_SCAT_SECTORS == 0)
	    case SCATTERED_IO:	r = do_vrdwt(&mess, wini_rdwt);	break;
#else
	    case SCATTERED_IO:	r = wini_hvrdwt(&mess);	break;
#endif
	    case DEV_IOCTL:	r = do_ioctl(&mess);	break;
	    default:		r = EINVAL;		break;
	}

	/* Finally, prepare and send the reply message. */
	mess.m_type = TASK_REPLY;	
	mess.REP_PROC_NR = procno;
	mess.REP_STATUS = r;	/* # of bytes transferred or error code */
	send(caller, &mess);	/* send reply to caller */
  }
}

/*===========================================================================*
 *				acsi_cmd				     *
 *===========================================================================*/
int acsi_cmd(drive, cmd, cmd_cnt, address, data_len, rw)
int drive;
unsigned char *cmd;
int cmd_cnt;
phys_bytes address;
phys_bytes data_len;
int rw;
{
  register int r = OK;
  int s;
  int wrbit;

  /*
   * Carry out the transfer. All parameters have been checked and
   * are set up properly.
   *
   * Every single byte written to the hdc will cause an interrupt.
   * Thus disable interrupts while communicating with hdc. Ready test
   * will be done by busy waiting. Only for real hard disk operations
   * interrupts will be enabled.
   */
  TRACE(printf("acsi_cmd: cmd: %lx, cmd_cnt: %d, data_len %ld, rw: %d\n",
	       cmd, cmd_cnt, data_len, rw));
  TRACE(printf("cmd: %d %d %d %d %d %d\n",
		cmd[0], cmd[1], cmd[2], cmd[3], cmd[4], cmd[5]));
  if (rw == DEV_READ) {
	wrbit = 0;
  } else {
	wrbit = WRBIT;
  }

#if (FASTDISK == 0)
  if (cmd_delay_enabled)
	cmd_delay();
  else
	cmd_delay_enabled = 1;
#endif

  dmagrab(WINCHESTER, hdcint);
  dmaaddr(address);		/* DMA address setup */
  IDISABLE();
  DMA->dma_mode = FDC | HDC;
  --cmd_cnt;
  dmawcmd((short) *cmd++, FDC | HDC | A0);
  DMA->dma_mode = FDC | HDC | A0;
  if (poll() != OK)
	 r = ERROR;
  while ((--cmd_cnt > 0) && (r == OK))	/* do not handle the last byte */
  {
    dmawcmd((short) *cmd++, FDC | HDC | A0);
    if (poll() != OK)
	r = ERROR;
  }
  IENABLE();
  if (r == OK)
  {
	/* fill the sector count register */
	dmacomm(wrbit | FDC | SCREG, (int)((data_len + 511)/512), HDC_DELAY);
	/* write the last command byte */
	dmawdat(wrbit | FDC | HDC | A0, *cmd, HDC_DELAY);
	DMA->dma_mode = wrbit;
	if (r == OK) {
		receive(HARDWARE, &mess);	/* Wait for interrupt. */
		s = dmardat(wrbit | FDC | HDC | A0, HDC_DELAY);
		if (s & HDC_CC) {
		    r = ERROR;
		}
	}
  }
  if (dmardat(wrbit | FDC | HDC | A0, HDC_DELAY))	/* finish HDC command */
	;
  if (dmardat(wrbit | HDC | A0, HDC_DELAY))
	;
  if ((data_len & 0xf) && (r == OK)) flush_dmaq();
  dmafree(WINCHESTER);
  return(r);
}

/*===========================================================================*
 *				flush_dmaq				     *
 *===========================================================================*/
PRIVATE void flush_dmaq()
{
/* we need to do request sense 4 times since due to buffering in
 * the dma controller will not transfer the data if it is less than
 * 16 bytes. request sense returns only four bytes...
 */
  register int i, r, s;
  unsigned char cmd[6];
  int cmd_cnt;

  cmd[0] = HD_RQS;
  cmd[1] =
  cmd[2] =
  cmd[3] =
  cmd[4] =
  cmd[5] = 0;

  for (i = 0; i < 4; i++) {

#if (FASTDISK == 0)
	if (cmd_delay_enabled)
		cmd_delay();
	else
		cmd_delay_enabled = 1;
#endif

	IDISABLE();
	DMA->dma_mode = FDC | HDC;
	cmd_cnt = 0;
	dmawcmd((short) cmd[0], FDC | HDC | A0);
	DMA->dma_mode = FDC | HDC | A0;
	if (poll() != OK)
		r = ERROR;
	while ((++cmd_cnt < 5) && (r == OK))	/* do not handle the last byte */
	{
	    dmawcmd((short) cmd[cmd_cnt], FDC | HDC | A0);
	    if (poll() != OK)
		r = ERROR;
	}
	IENABLE();
	if (r == OK)
	{
		/* fill the sector count register */
		dmawdat(FDC | SCREG, 1, HDC_DELAY);
		/* write the last command byte */
		dmawdat(FDC | HDC | A0, cmd[5], HDC_DELAY);
		DMA->dma_mode = 0;
		if (r == OK) {
			receive(HARDWARE, &mess);	/* Wait for interrupt. */
			s = dmardat(FDC | HDC | A0, HDC_DELAY);
			if (s & HDC_CC) {
			    r = ERROR;
			}
		}
	}
  if (dmardat(FDC | HDC | A0, HDC_DELAY))	 /* finish HDC command */
	;
  if (dmardat(HDC | A0, HDC_DELAY))
	;
  }
}

/*===========================================================================*
 *				poll					     *
 *===========================================================================*/
PRIVATE int poll()
{
  register int	i;

  /* wait until hdc signals ready - or return ERROR after t tries */
  for (i = HDC_POLL; --i >= 0; )
	if ((MFP->mf_gpip & IO_DINT) == 0)
		return(OK);
  return(ERROR);
}

/*===========================================================================*
 *				hdcint					     *
 *===========================================================================*/

PRIVATE void hdcint()
{
  interrupt(WINCHESTER);
}

/*===========================================================================*
 *				do_ioctl				     *
 *===========================================================================*/
PRIVATE int do_ioctl(mp)
register message *mp;
{
  register phys_bytes	uaddress;
  register struct proc	*rp;
  register int r;

  rp = proc_addr(mp->PROC_NR);
  switch(mp->TTY_REQUEST) {
#if (HD_CLOCK == 1)
     case DC_RBMS100:
     case DC_RBMS200:
     case DC_RSUPRA:
     case DC_RICD:
	  uaddress = umap(rp, D, (vir_bytes) mp->ADDRESS, (vir_bytes) XFERSIZE);
	  if (uaddress == 0)
		return(EINVAL);
	  r = do_xbms(uaddress,XFERSIZE,DEV_READ,(int)mp->TTY_REQUEST);
	  break;
     case DC_WBMS100:
     case DC_WBMS200:
	  uaddress = umap(rp, D, (vir_bytes) mp->ADDRESS, (vir_bytes) XFERSIZE);
	  if (uaddress == 0)
		return(EINVAL);
	  r = do_xbms(uaddress,XFERSIZE,DEV_WRITE,(int)mp->TTY_REQUEST);
	  break;
#endif /* HD_CLOCK */
     default:
	  r = wini_ioctl(mp);
  }
  return(r);
}
#endif 
#endif 
