/*
 * This file contains a driver for the Floppy Disk Controller (FDC)
 * on the Atari ST. It uses the WD 1772 chip, modified for steprates.
 * The driver supports two operations: read a block and write a block.
 * It accepts messages for reading, for writing, for scattered I/O, for
 * control and for opening and closing a device. All use message format
 * m2 with the following parameters:
 *
 *    m_type      DEVICE    PROC_NR    COUNT   POSITION   ADRRESS  TTY_FLAGS
 * --------------------------------------------------------------------------
 * |  DEV_READ  | device  | proc nr |  bytes  |  offset | buf ptr |         |
 * |------------+---------+---------+---------+---------+---------+---------|
 * |  DEV_WRITE | device  | proc nr |  bytes  |  offset | buf ptr |         |
 * |------------+---------+---------+---------+---------+---------+---------|
 * |SCATTERED_IO| device  | proc nr | requests|         | iov ptr |         |
 * |------------+---------+---------+---------+---------+---------+---------|
 * |  DEV_IOCTL | device  | proc nr |func code|         | arg ptr |         |
 * |------------+---------+---------+---------+---------+---------+---------|
 * |  DEV_OPEN  | device  | proc nr |         |         |         |open mode|
 * |------------+---------+---------+---------+---------+---------+---------|
 * |  DEV_CLOSE | device  | proc nr | #links  |         |         |         |
 * --------------------------------------------------------------------------
 *
 * The file contains two entry points:
 *
 *	floppy_task:	main entry when system is brought up
 *	fd_timer:	used to deselect drives, called from clock.c
 *
 * The ST floppy disk controller shares the access to the DMA circuitry
 * with other devices. For this reason the floppy disk controller makes
 * use of some special DMA accessing code.
 *
 * Interrupts from the FDC are in fact DMA interrupts which get their
 * first level handling in stdma.c . If the floppy driver is currently
 * using DMA the interrupt is signalled to the floppy task.
 *
 * For the MegaSTE and the TT HD-Drives are interfaced via DMA->fdc_mode
 * (0xFFFF860EL). For normal STs PA_USER and PA_FDDENS are used for
 * density selection. (/PA_USER & PA_FDDENS) should be connected to the
 * Shugartbus signal HD-IN (Pin 2).
 * The following frequency assignment is used:
 *	PA_USER PA_FDDENS	Frequency	HD-IN
 *	0	0 		 8    Mhz	0
 *	0	1		16    Mhz	1
 *	1	0		10.24 Mhz	0
 *	1	1		 9.6  Mhz	0
 *
 *
 * The minor device number is build as follows
 *	0TTTTT0N
 *	N	= drive nr.
 *	TTTTT	= disktype
 *
 * Supported ioctl calls:
 *	DIOGETP    - ioctl(fd, DIOGETP, (struct fparam *)p)
 *			The struct fparam pointed to by p is filled
 *			with the disktype's format description.
 *			For autoconfig devices this description
 *			will be initialized at open time.
 *	DIOSETP    - ioctl(fd, DIOSETP, (struct fparam *)p)
 *			This call is only legal for disktypes 0 and 16.
 *			The kernel fparam field associated with fd will
 *			be filled with the user supplied structure pointed
 *			to by p. For autoconfig devices (disktype 0) the
 *			structure will be reset on open.
 *	DIOGETHP   - ioctl(fd, DIOGETHP, (struct hparam *)p)
 *			The struct hparam pointed to by p is filled with
 *			the hardware parameters of the drive associated with
 *			fd.
 *	DIOSETHP   - ioctl(fd, DIOSETHP, (struct hparam *)p)
 *			The kernel hparam field for the drive associated with
 *			fd will be filled with the user supplied structure
 *			pointed to by p.
 *	DIOFMTLOCK - ioctl(fd, DIOFMTLOCK, NULL)
 *			This call will only succeed if fd is the only
 *			open fd for the drive and will prevent further
 *			opens for the drive. After issuing DIOFMTLOCK
 *			DIOSETP/DIOSETHP will always fail with EBUSY.
 *			The data returned by DIOGETP will be modified in
 *			the following way:
 *			- nr_sectors will be set to 1
 *			- sector_0 will be set to 1
 *			- sector_size will be set to the tracksize rounded
 *			  up to the next 512 byte boundary.
 *	DIOFMTFREE - ioctl(fd, DIOFMTFREE, NULL)
 *			This call will reverse the effect of DIOFMTLOCK.
 *
 * Under some conditions the autoconfig code will fail!
 *		- Some TOS-Formatters (and an early version of fdfmt.c :-(... )
 *		  will write dummy sector headers at the beginning of each
 *		  track to go around the performance problems issued by the
 *		  verify-after-seek that is done by the Atari ROM-BIOS
 *		  (note that there are good reasons for the verify-after-seek...).
 *		  These additional sector headers will confuse the autoconfig
 *		  code. If the formatter e.g. inserts sector headers with the
 *		  sector header set to 0, autoconfig will think that the first
 *		  sector number on each track is 0.
 *
 *		- There are still several formats not handled by this driver:
 *		  o Formats with different sector numbering for side 0 and 1 e.g.:
 *		    Side 0:  1  2  3  4  5  6  7  8  9
 *		    Side 1: 10 11 12 13 14 15 16 17 18
 *		  o Formats with nonconsecutive sector numbering scheme e.g.:
 *		    1 4 5 7 10 15 ...
 *		  o Formats with nonconsecutive track numbers.
 *		  o Formats with logical sector translation.
 *		  o Formats with "deleted data marks".
 *		  o Formats with different densities on different tracks
 *		    (e.g. ECMA70).
 *		  o FM Formats.
 *		  o Formats I could not imagine :-)
 *		  You can see there is still enough to do for you...
 *
 * For more information read:
 *	<sys/fdio.h>
 *	commands/atari/fdfmt.c
 *	commands/atari/fdctrl.c
 *		11'92 Matthias Pfaller (leo @ marco.de)
 */

#include "kernel.h"
#if (MACHINE == ATARI)
#if (NR_FD_DRIVES == 0)
PUBLIC void fd_timer()
{
}
#else /* (NR_FD_DRIVES > 0) */
#include <minix/callnr.h>
#include <minix/com.h>
#include <sgtty.h>
#include <sys/fdio.h>
#include <fcntl.h>

#include "proc.h"

#include "staddr.h"
#include "stfdc.h"
#include "stdma.h"
#include "stsound.h"

#ifdef _ANSI
#define ASSERT(x)       if (!(x)) panic("fd: ASSERT(" #x ") failed", NO_NUM)
#else
#define ASSERT(x)       if (!(x)) panic("fd: ASSERT(x) failed", NO_NUM)
#endif
#define TRACE(x)	/* x */
#define DEBUG(x)	x

#define	FDC_DELAY	32	/* for dma[rw]dat() */

/* Parameters for the disk drive. */
#define NR_FLOPDRIVES	2	/* maximum number of drives */
#define MAX_ERRORS	10	/* how often to try rd/wt before quitting */

#define NR_TYPES	((sizeof(fparam)/(sizeof(struct fparam *) * 2)) - 1)
				/* number of formats */
#define TRACK_SIZE	12800	/* max size of unformatted disk */

/*
 * tracklen in bytes for the different densities
 */
		      /* DD    QD    HD5    HD3 */
PRIVATE int trklen[] = { 6656, 8192, 10752, 12800 };

/*
 * DRIVE returns the select code
 */
#define	DRIVE(d)	((d) & 0x01)

/*
 * DTYPE returns the disk type number
 */
#define	DTYPE(d)	(((d) >> 2) & 0x1F)

/*
 * structure to count how many times a drive is open
 */
PRIVATE struct open_count {
	int o_cnt;
	int formatting;
} open_count[NR_FLOPDRIVES] = {0};

/*
 * drive parameters
 */
#if ((ATARI_TYPE == TT) || (ATARI_TYPE == MSTE))
PRIVATE struct hparam hparam[] = {
	{ 80, (1<<DD)|(1<<HD), I3, SIDES2, FAST },
	{ 80, (1<<DD)|(1<<HD), I3, SIDES2, FAST },
};
#else
PRIVATE struct hparam hparam[] = {
	{ 80, (1<<DD)|(1<<QD)|(1<<HD), I3, SIDES2, FAST },
	{ 80, (1<<DD)|(1<<QD)|(1<<HD), I3, SIDES2, FAST },
};
#endif

/*
 * variable parameters for each minor device
 */
PRIVATE struct fparam fparams[] = {
{  512, 1, 10,	0, 80, SIDES2, DD,  NSTP, AUTO }, /* Autoconfig drive A    0 */
{  512, 1, 10,	0, 80, SIDES2, DD,  NSTP, AUTO }, /* Autoconfig drive B    1 */
{  512, 1, 10,  0, 80, SIDES2, DD,  NSTP, HARD }, /* Configurable drive A  2 */
{  512, 1, 10,	0, 80, SIDES2, DD,  NSTP, HARD }, /* Configurable drive B  3 */
{  512, 1,  9,	0, 40, SIDES2, DD,  NSTP, HARD }, /* 5.25'', 360k, 40 trk  4 */
{  512, 1, 15,	0, 80, SIDES2, HD,  NSTP, HARD }, /* 5.25'', 1.2M, 80 trk  5 */
{  512, 1,  9,	0, 40, SIDES2, DD,  DSTP, HARD }, /* 5.25'', 360k, 80 trk  6 */
{  512, 1,  9,	0, 80, SIDES1, DD,  NSTP, HARD }, /* 3.5'',  360k, 80 trk  7 */
{  512, 1,  9,	0, 80, SIDES2, DD,  NSTP, HARD }, /* 3.5'',  720k, 80 trk  8 */
{  512, 1,  9,	0, 40, SIDES2, DD5, DSTP, HARD }, /* 5.25'', 360k, 80 trk  9 */
{  512, 1,  9,	0, 80, SIDES2, DD5, NSTP, HARD }, /* 5.25'', 720k, 80 trk 10 */
{  512, 1, 18,	0, 80, SIDES2, HD,  NSTP, HARD }, /* 3.5'', 1.44M, 80 trk 11 */
{  512, 1, 10, 15, 65, SIDES2, DD,  NSTP, HARD }, /* boot disk		  12 */
{  512, 1, 13,	0, 80, SIDES2, QD,  NSTP, HARD }, /* 3.5'', 1.04M, 80 trk 13 */
{ 1024, 1,  7,	0, 80, SIDES2, QD,  NSTP, HARD }, /* 3.5'', 1.11M, 80 trk 14 */
{ 1024, 1, 10,	0, 80, SIDES2, HD,  NSTP, HARD }, /* 3.5'',  1.6M, 80 trk 15 */
{ 1024, 1, 11,	0, 80, SIDES2, HD,  NSTP, HARD }, /* 3.5'',  1.7M, 80 trk 16 */
};

PRIVATE struct fparam format;	/* fparam struct used for formatting */

PRIVATE struct fparam *fparam[][2] = { /* index */	/* minor */
  { &fparams[ 0], &fparams[ 1] },	/*  0 */ 	/*  0,  1 */
  { &fparams[ 4], &fparams[ 4] },	/*  1 */ 	/*  4,  5 */
  { &fparams[ 5], &fparams[ 5] },	/*  2 */ 	/*  8,  9 */
  { &fparams[ 6], &fparams[ 6] },	/*  3 */ 	/* 12, 13 */
  { &fparams[ 8], &fparams[ 8] },	/*  4 */ 	/* 16, 17 */
  { &fparams[ 9], &fparams[ 9] },	/*  5 */ 	/* 20, 21 */
  { &fparams[10], &fparams[10] },	/*  6 */ 	/* 24, 25 */
  { &fparams[11], &fparams[11] },	/*  7 */ 	/* 28, 29 */
  { &fparams[ 7], &fparams[ 7] },	/*  8 */ 	/* 32, 33 */
  { NULL,	  NULL	       },	/*  9 */ 	/* 36, 37 */
  { NULL,	  NULL	       },	/*  a */ 	/* 40, 41 */
  { NULL,	  NULL	       },	/*  b */ 	/* 44, 45 */
  { NULL,	  NULL	       },	/*  c */ 	/* 48, 49 */
  { NULL,	  NULL	       },	/*  d */ 	/* 52, 53 */
  { NULL,	  NULL	       },	/*  e */ 	/* 56, 57 */
  { NULL,	  NULL	       },	/*  f */ 	/* 60, 61 */
  { &fparams[ 2], &fparams[ 3] },	/* 10 */ 	/* 64, 65 */
  { &fparams[12], &fparams[12] },	/* 11 */ 	/* 68, 69 */
  { &fparams[13], &fparams[13] },	/* 12 */ 	/* 72, 73 */
  { &fparams[14], &fparams[14] },	/* 13 */ 	/* 76, 77 */
  { &fparams[15], &fparams[15] },	/* 14 */ 	/* 80, 81 */
  { &fparams[16], &fparams[16] },	/* 15 */ 	/* 84, 85 */
};

/* return values of xfer_ok(): */
#define X_OK		0
#define X_AGAIN		1
#define X_ERROR		2
#define X_FAIL		3

PRIVATE int curcyl[NR_FLOPDRIVES];	/* current cylinder (-1 if not calibrated) */

PRIVATE message mess;		/* message buffer for in and out */

PRIVATE struct xfer {
	int	x_rw;		/* read or write */
	int	x_drive;	/* drive number */
	int	x_secnum;	/* current sector */
	phys_bytes x_address;	/* current physical address */
	int	x_count;	/* bytes still to transfer */
	int	x_errors;	/* errors on current sector */
	int	x_cmd;		/* controller command */
} xfer;

FORWARD _PROTOTYPE( int do_open, (message *mp)				);
FORWARD _PROTOTYPE( int do_close, (message *mp)				);
FORWARD _PROTOTYPE( int do_rdwt, (message *mp)				);
FORWARD _PROTOTYPE( int do_ioctl, (message *mp)				);
FORWARD _PROTOTYPE( int do_xfer, (void)					);
FORWARD _PROTOTYPE( int fd_select, (int drive, int head, int dense)	);
FORWARD _PROTOTYPE( void fd_deselect, (void)				);
FORWARD _PROTOTYPE( void fdcint, (void)					);
FORWARD _PROTOTYPE( int xfer_done, (void)				);
FORWARD _PROTOTYPE( int xfer_ok, (void)					);
FORWARD _PROTOTYPE( struct fparam *gparam, (int drive)			);
FORWARD _PROTOTYPE( void sparam, (int drive, struct fparam *f)		);

/*===========================================================================*
 *				floppy_task				     *
 *===========================================================================*/
PUBLIC void floppy_task()
{
  register int	r, drive, caller, procno;

  /*
   * The main loop of the disk task.
   * It waits for a message, carries it out, and sends a reply.
   */
  TRACE(printf("fd: task started\n"));

  for (drive = 0; drive < NR_FLOPDRIVES; drive++)
	curcyl[drive] = -1;	/* uncalibrated */

  while (TRUE) {
	receive(ANY, &mess);
	ASSERT(mess.m_source >= 0);
	TRACE(printf("fd: received %d from %d\n", mess.m_type,mess.m_source));
	caller = mess.m_source;
	procno = mess.PROC_NR;

	/* Now carry out the work. */
	switch (mess.m_type) {
	    case DEV_OPEN:	r = do_open(&mess);	break;
	    case DEV_CLOSE:	r = do_close(&mess);	break;
	    case DEV_READ:
	    case DEV_WRITE:	r = do_rdwt(&mess);	break;
	    case SCATTERED_IO:	r = do_vrdwt(&mess, do_rdwt); break;
	    case DEV_IOCTL:	r = do_ioctl(&mess);	break;
	    default:		r = EINVAL;		break;
	}

	/* Finally, prepare and send the reply message. */
	mess.m_type = TASK_REPLY;	
	mess.REP_PROC_NR = procno;
	mess.REP_STATUS = r;	/* # of bytes transferred or error code */
	send(caller, &mess);	/* send reply to caller */
  }
}


/*===========================================================================*
 *				do_open					     *
 *===========================================================================*/
#define reset_xfer(xp, start) \
	xp->x_secnum  = start;\
	xp->x_address = kaddress;\
	xp->x_count   = 3 * sizeof(struct idmark);\
	xp->x_errors  = 0;

PRIVATE int do_open(mp)
register message *mp;
{
  register struct fparam *f;
  register struct hparam *h;
  register struct xfer *xp;
  register int r = OK, sec1, secmin, secmax, sec0;
  phys_bytes kaddress;
  struct idmark ids[3];

  if (DRIVE(mp->DEVICE) >= NR_FLOPDRIVES)
	return(ENXIO);
  if (DTYPE(mp->DEVICE) >= NR_TYPES)
	return(ENXIO);
  if (open_count[DRIVE(mp->DEVICE)].formatting)
	return(EBUSY);
  if (!(f = gparam(mp->DEVICE)))
	return(ENXIO);
  h = &hparam[DRIVE(mp->DEVICE)];

  dmagrab(FLOPPY, fdcint);
  fd_select(DRIVE(mp->DEVICE), 0, DD);

  /* This updates the WRI_PRO bit */
  dmawdat(FDC_CS, IRUPT, FDC_DELAY);
  if ((dmardat(FDC_CS, FDC_DELAY) & WRI_PRO) &&
      ((mp->TTY_FLAGS & 3) != O_RDONLY))
	r = EACCES;
  else
	r = OK;

  if (r == OK && f->autocf == AUTO) {
	kaddress = umap(proc_ptr, S, (vir_bytes) ids, (vir_bytes) sizeof(ids));
	f->nr_sides = h->sides;

	/*
	 * sec0 is set to a value that will result in a
	 * seek to track 2 to test for double stepping.
	 */
	sec0 = f->nr_sectors * 2 * (f->nr_sides?2:1);
	xp            = &xfer;
	xp->x_rw      = mp->m_type;
	xp->x_drive   = mp->DEVICE;

	/* Probe density */
	for (r = EIO, f->density = DD; f->density <= DD5; f->density++) {
		if (!(h->dense & (1<<f->density)))
			continue;
		reset_xfer(xp, sec0);
		if (do_xfer())
			receive(HARDWARE, &mess);
		else
			xp->x_errors = MAX_ERRORS;
		if (xp->x_errors < MAX_ERRORS) {
			r = OK;
			break;
		}
	}
	if (r != OK) {
		fd_deselect();
		dmafree(FLOPPY);
		return(r);
	}

	/* Probe number of sectors/track */
	secmin = secmax = sec1 = ids[0].secnum;
	if (ids[1].secnum > sec1)
		secmax = ids[1].secnum;
	else
		secmin = ids[1].secnum;
	if (ids[2].secnum > secmax) secmax = ids[2].secnum; else
	if (ids[2].secnum < secmin) secmin = ids[2].secnum;
	do {
		reset_xfer(xp, sec0);
		if (do_xfer())
			receive(HARDWARE, &mess);
		else
			xp->x_errors = MAX_ERRORS;
		if (xp->x_errors >= MAX_ERRORS) {
			fd_deselect();
			dmafree(FLOPPY);
			return(EIO);
		}

		if (ids[0].secnum == sec1) break;
		if (ids[0].secnum > secmax) secmax = ids[0].secnum; else
		if (ids[0].secnum < secmin) secmin = ids[0].secnum;

		if (ids[1].secnum == sec1) break;
		if (ids[1].secnum > secmax) secmax = ids[1].secnum; else
		if (ids[1].secnum < secmin) secmin = ids[1].secnum;

		if (ids[2].secnum == sec1) break;
		if (ids[2].secnum > secmax) secmax = ids[2].secnum; else
		if (ids[2].secnum < secmin) secmin = ids[2].secnum;
	} while (1);
	f->sector_0   = secmin;
	f->nr_sectors = secmax - secmin + 1;

	/* set sector size */
	f->sector_size = 128 << ids[0].seclen;

	/* Test for double stepping */
	if (ids[0].trknum == 1) {
		f->stepping = DSTP;
		f->nr_cylinders = h->max_cyl / 2;
	} else {
		f->stepping = NSTP;
		f->nr_cylinders = h->max_cyl;
	}		
	f->cylinder_0 = 0;

	if (h->sides == SIDES2) {
		reset_xfer(xp, sec0 + f->nr_sectors);
		if (do_xfer())
			receive(HARDWARE, &mess);
		else
			r = EIO;
		if (xp->x_errors >= MAX_ERRORS)
			f->nr_sides = SIDES1;
		else
			f->nr_sides = SIDES2;
	}
  }
  fd_deselect();
  dmafree(FLOPPY);
  if (r == OK)
	open_count[DRIVE(mp->DEVICE)].o_cnt++;
  return(r);
}


/*===========================================================================*
 *				do_close				     *
 *===========================================================================*/
PRIVATE int do_close(mp)
register message *mp;
{
  if (mp->COUNT != 0)
	return(OK);

  if (open_count[DRIVE(mp->DEVICE)].o_cnt == 0) {
	printf("FD: MINOR %d: Closing non open device\n", mp->DEVICE);
	return(OK);
  }
  if (!--open_count[DRIVE(mp->DEVICE)].o_cnt)
	open_count[DRIVE(mp->DEVICE)].formatting = 0;

  return(OK);
}


/*===========================================================================*
 *				do_rdwt					     *
 *===========================================================================*/
PRIVATE int do_rdwt(mp)
register message *mp;
{
  register struct xfer *xp;
  register struct fparam *f;
  register struct proc *rp;
  register int	nbytes;
  phys_bytes dst_addr;
  int max_bytes;
#if (ATARI_TYPE == TT)
  static char buf[TRACK_SIZE]; /* buffer to make sure we dma to/from ST ram  */
			       /* the buffer must be large enough to contain */
			       /* a complete unformatted track		*/
#endif

  xp = &xfer;
  xp->x_rw = mp->m_type;
  xp->x_drive = mp->DEVICE;
  if (DRIVE(xp->x_drive) >= NR_FLOPDRIVES)
	return(ENXIO);
  if (DTYPE(xp->x_drive) >= NR_TYPES)
	return(ENXIO);
  if (open_count[DRIVE(xp->x_drive)].formatting && xp->x_rw == DEV_READ)
	return(EIO);
  if (!(f = gparam(xp->x_drive)))
	return(ENXIO);

  nbytes = mp->COUNT;

  if ((mp->POSITION % f->sector_size) != 0)
	return(EINVAL);
  if ((nbytes % f->sector_size) != 0)
	return(EINVAL);
  xp->x_secnum = (int)(mp->POSITION/f->sector_size);
  rp = proc_addr(mp->PROC_NR);
  dst_addr = umap(rp, D, (vir_bytes) mp->ADDRESS, (vir_bytes) nbytes);
  if (dst_addr == 0)
	return(EINVAL);
  if (nbytes == 0)
	return(0);
#if (ATARI_TYPE == TT)
  xp->x_address = umap(proc_ptr, D, (vir_bytes) buf, (vir_bytes) TRACK_SIZE);
  max_bytes = TRACK_SIZE;
#else
  max_bytes = nbytes;
  xp->x_address = dst_addr;
#endif
  do {
#if (ATARI_TYPE == TT)
	if (xp->x_rw == DEV_WRITE)
		phys_copy(dst_addr, xp->x_address, (vir_bytes) max_bytes);
#endif
	if (nbytes <= max_bytes) {
		max_bytes = nbytes;
	}
	xp->x_count = max_bytes;
	nbytes -= max_bytes;
	rp->p_physio = 1;		/* disable (un)shadowing */
	dmagrab(FLOPPY, fdcint);
	xp->x_errors = 0;
	if (do_xfer()) {
		receive(HARDWARE, &mess);
		fd_deselect();
	}
	dmafree(FLOPPY);
	rp->p_physio = 0;		/* enable (un)shadowing */
	if (xp->x_errors >= MAX_ERRORS)
  		return(EIO);
	if (xp->x_count != 0)
		return(EIO);		/* partial transfer */
#if (ATARI_TYPE == TT)
	xp->x_address -= max_bytes;	/* set pointer back */
	if (xp->x_rw == DEV_READ)
		phys_copy(xp->x_address, dst_addr, (vir_bytes) max_bytes);
	dst_addr += max_bytes;
#endif
  } while (nbytes > 0);
  return(mp->COUNT);
}

/*===========================================================================*
 *				do_xfer					     *
 *===========================================================================*/
PRIVATE int do_xfer()
{
  register struct xfer *xp;
  register struct fparam *f;
  register struct hparam *h;
  register int d, head, cylinder, sector, hbit, dense, seekrate;
  static int ebit = 0;

  xp = &xfer;

  f = gparam(xp->x_drive);
  cylinder = xp->x_secnum / f->nr_sectors;
  if (f->nr_sides == SIDES2) {
	head = cylinder & 1;
	cylinder >>= 1;
  } else
	head = 0;

  if (f->stepping == DSTP)
	cylinder <<= 1;
  cylinder += f->cylinder_0;
  if (cylinder >= (f->cylinder_0 + f->nr_cylinders))
	return(0);

  dense = f->density;
  d = DRIVE(xp->x_drive);
  h = &hparam[d];
  hbit = 0;
  if (fd_select(d, head, dense) != 0)
	hbit = HBIT;	/* motor on, suppress spin up sequence */
  if (dense != DD)
	seekrate = (h->slow == SLOW)?RATE12:RATE6;
  else
	seekrate = (h->slow == SLOW)?RATE6:RATE3;

  switch (xp->x_rw) {
	case DEV_READ:
		xp->x_cmd = F_READ;
		break;
	case DEV_WRITE:
		xp->x_cmd = open_count[d].formatting ? WRITETR:F_WRITE;
		break;
	case DEV_OPEN:
		xp->x_cmd = READID;
		break;
	default:
		printf("\nKernel panic: FD: MINOR %d: bad xp->x_rw(%d)\n", xp->x_drive, xp->x_rw);
		panic(NULL, NO_NUM);
		return(0);
  }
  
  if (curcyl[d] == -1) {
	/* 
	 * Recalibrate, since we lost track of head positioning.
	 * The floppy disk controller has no way of determining its
	 * absolute arm position (cylinder).  Instead, it steps the
	 * arm a cylinder at a time and keeps track of where it
	 * thinks it is (in software).  However, after a SEEK, the
	 * hardware reads information from the diskette telling
	 * where the arm actually is.  If the arm is in the wrong place,
	 * a recalibration is done, which forces the arm to cylinder 0.
	 * This way the controller can get back into sync with reality.
	 */
	TRACE(printf("fd%d: recalibrate\n", xp->x_drive));
	xp->x_cmd = RESTORE;
	dmawdat(FDC_CS, RESTORE|hbit|seekrate, FDC_DELAY);
	ebit = EBIT;
	return(1);
  }

  dmawdat(FDC_TR, curcyl[d], FDC_DELAY);

  /*
   * Issue a SEEK command on the indicated drive unless the arm is
   * already positioned on the correct cylinder.
   */
  if (cylinder != curcyl[d]) {
	curcyl[d] = cylinder;	/* be optimistic */
	xp->x_cmd = SEEK;
	dmawdat(FDC_DR, cylinder, FDC_DELAY);
	dmawdat(FDC_CS, SEEK|hbit|seekrate, FDC_DELAY);
	ebit = EBIT;
	return(1);
  }

  /* The drive is now on the proper cylinder.  Read or write 1 block. */
  sector = (xp->x_secnum % f->nr_sectors) + f->sector_0;

  TRACE(printf("fd%d: %s: secnum=%d,cylinder=%d,sector=%d,head=%d\n",
	xp->x_drive, xp->x_rw == DEV_READ ? "read" : "write",
	xp->x_secnum, cylinder, sector, head));

  dmawdat(FDC_SR, sector, FDC_DELAY);
  dmaaddr(xp->x_address);	/* DMA address setup */

  if (xp->x_rw == DEV_READ || xp->x_rw == DEV_OPEN) {
	/* Issue the command */
	dmacomm(FDC_CS | SCREG, (f->sector_size + 511) / 512, 0);
	dmawdat(FDC_CS, xp->x_cmd|hbit|ebit, FDC_DELAY);
  } else {
	/* Issue the command */
	dmacomm(WRBIT | FDC_CS | SCREG, (f->sector_size + 511) / 512, FDC_DELAY);
	dmawdat(WRBIT | FDC_CS, xp->x_cmd|hbit|ebit, FDC_DELAY);
  }
  ebit = 0;
  return(1);
}

/*===========================================================================*
 *			fd_select, fd_deselect, fd_timer		     *
 *===========================================================================*/

PRIVATE int selected;		/* drive/head currently selected */
PRIVATE int selectic = 30;	/* # of fd_timer calls to deselect */
PRIVATE int motoron;		/* for how long will motor run? */

PRIVATE int fd_select(drive, head, dense)
int drive;
int head;
int dense;
{
  register int	i, s, spinning;

  TRACE(printf("fd: select (drive %d, head %d)\n", drive, head));
  spinning = motoron;
  motoron = 5;		/* will run for more than 500 msec */
  selectic = -30;	/* error if not ready within 3 sec */
  i = (drive == 1 ? PA_FLOP1 : PA_FLOP0);
  i |= head;
  switch(dense)
  {
#if ((ATARI_TYPE == TT) || (ATARI_TYPE == MSTE))
	case HD:
		DMA->fdc_mode = FD_HD;
		break;

	case DD5:
		/* No DD5 - would a panic be better? */
	case QD:
		/* No QD  - would a panic be better? */
	case DD:
		DMA->fdc_mode = FD_DD;
		break;
#else
	case HD:
		i |= PA_FDDENS;
		break;

	case DD5:
		i |= PA_FDDENS | PA_USER;
		break;

	case QD:
		i |= PA_USER;
		break;

	case DD:
		break;
#endif

	default:
		printf("\nKernel panic: FD: MINOR %d: illegal density code %d\n", xfer.x_drive, dense);
		panic(NULL, NO_NUM);
		break;
  }

  if (i != selected) {
	selected = i;
	s = lock();
	SOUND->sd_selr = YM_IOA;
	SOUND->sd_wdat = (SOUND->sd_rdat & 0x78) | (i ^ 0x07);
	restore(s);
  }
  return(spinning);
}

PRIVATE void fd_deselect()
{
  TRACE(printf("fd: deselect\n"));
  selectic = 30;	/* postpone for 30 * 100 msec == 3 sec */
}

PUBLIC void fd_timer()
{
/*
 * Perform three functions:
 * 1. Estimate when motor stops running
 * 2. Force interrupt if drive does not respond
 * 3. Deselect drive(s) when all motors stopped spinning
 */
  register int	i, pl;

  if (selectic == 0)
	return;
  if (motoron)
	--motoron;
  if (selectic < 0) {
	if (++selectic == 0) {
		pl = lock();
		xfer.x_cmd = IRUPT;
		/*
		 * IRUPT|IFORCE should be used. This does not
		 * work as expected for unknown reasons.
		 * RESTORE should never fail and will force
		 * an interrupt too.
		 */
		dmawdat(FDC_CS, IRUPT, FDC_DELAY);
		for (i = 0; i < 256; i++)
			;
		dmardat(FDC_CS, FDC_DELAY);
		dmawdat(FDC_CS, RESTORE, FDC_DELAY);
		restore(pl);
	}
	return;
  }
  if (--selectic != 0)
	return;
  selected = 0;
  pl = lock();
  SOUND->sd_selr = YM_IOA;
  /* It is better to switch back to 8 Mhz
   * when the FDC is not in use. This will
   * reduce heating.
   */
#if ((ATARI_TYPE == TT) || (ATARI_TYPE == MSTE))
  SOUND->sd_wdat = (SOUND->sd_rdat | 0x07);
  DMA->fdc_mode = FD_DD;
#else
  SOUND->sd_wdat = (SOUND->sd_rdat | 0x07) & ~(PA_FDDENS|PA_USER);
#endif
  restore(pl);
}

/*===========================================================================*
 *				fdcint					     *
 *===========================================================================*/
PRIVATE void fdcint()
{
  ASSERT(xfer.x_count);
  if (xfer_done() == 0) {
	return;
  }
  interrupt(FLOPPY);
}

/*===========================================================================*
 *				xfer_done				     *
 *===========================================================================*/
PRIVATE int xfer_done()
{
  register struct xfer *xp;
  register struct fparam *f;

  xp = &xfer;
  f = gparam(xp->x_drive);
  switch (xfer_ok()) {
  case X_FAIL:
	return(1);
  case X_ERROR:
	xp->x_errors++;
	if (xp->x_errors >= MAX_ERRORS)
		return(1);
	break;
  case X_AGAIN:
	break;
  case X_OK:
	xp->x_errors = 0;
	xp->x_secnum++;
	if (xp->x_cmd == READID) {
		xp->x_address += sizeof(struct idmark);
		xp->x_count   -= sizeof(struct idmark);
		if (xp->x_count <= 0)
			return(1);
		dmawdat(FDC_CS, xp->x_cmd, FDC_DELAY);
		return(0);
	} else {
		xp->x_address += f->sector_size;
		xp->x_count   -= f->sector_size;
		if (xp->x_count <= 0)
			return(1);
	}
	break;
  }
  if (do_xfer() == 0)
	return(1);
  return(0);
}

/*===========================================================================*
 *				xfer_ok					     *
 *===========================================================================*/
PRIVATE int xfer_ok()
{
  register struct xfer *xp;
  register int	status;

  xp = &xfer;
  switch (xp->x_cmd) {
  case IRUPT:
	status = dmardat(FDC_CS, FDC_DELAY);
	if (xp->x_rw == DEV_OPEN) {
		xp->x_errors = MAX_ERRORS - 1;
	}
	printf("FD: MINOR %d: timeout\n", xp->x_drive);
	curcyl[DRIVE(xp->x_drive)] = -1;
	return(X_ERROR);
  case F_READ:
  case READID:
	/* Test for DMA error */
	status = dmastat(FDC_CS | SCREG, 0);
	if ((status & DMAOK) == 0) {
		printf("FD: MINOR %d: READ: dma status = 0x%x\n",
			xp->x_drive, DMA->dma_stat & BYTE);
		return(X_ERROR);
	}
	/* Get controller status and check for errors. */
	status = dmardat(FDC_CS, FDC_DELAY);
	if (status & (RNF | CRCERR | LD_T00)) {
		if (xp->x_cmd == READID) {
			xp->x_errors = MAX_ERRORS - 1;
		} else {
			printf("FD: MINOR %d: READ: sector=%d fdc status=0x%x\n",
				xp->x_drive, xp->x_secnum, status & 0xFF);
			if (xp->x_errors > 2)
				curcyl[DRIVE(xp->x_drive)] = -1;
		}
		return(X_ERROR);
	}
	break;
  case F_WRITE:
  case WRITETR:
	/* Test for DMA error */
	status = dmastat(FDC_CS | SCREG, 0);
	if ((status & DMAOK) == 0) {
		printf("FD: MINOR %d: WRITE: dma status = 0x%x\n",
			xp->x_drive, DMA->dma_stat & BYTE);
		return(X_ERROR);
	}
	/* Get controller status and check for errors. */
	status = dmardat(WRBIT | FDC_CS, FDC_DELAY);
	if (status & WRI_PRO) {
		printf("FD: MINOR %d: write protected\n", xp->x_drive);
		return(X_FAIL);
	}
	if (status & (RNF | CRCERR | LD_T00)) {
		printf("FD: MINOR %d: WRITE: sector=%d fdc status=0x%x\n",
			xp->x_drive, xp->x_secnum, status&0xFF);
		/* we retry two times in without recalibrate
		   to get softerrors without seeking */
		if (xp->x_errors > 2)
			curcyl[DRIVE(xp->x_drive)] = -1;
		return(X_ERROR);
	}
	break;
  case SEEK:
	status = dmardat(FDC_CS, FDC_DELAY);
	if (status & (RNF | CRCERR)) {
		curcyl[DRIVE(xp->x_drive)] = -1;
		return(X_ERROR);
	}
	return(X_AGAIN);
  case RESTORE:
	/* Determine if the recalibration succeeded. */
	status = dmardat(FDC_CS, FDC_DELAY);
	if (status & RNF) {
		printf("FD: MINOR %d: recalibrate failed fdc status=0x%x\n",
				xp->x_drive, status&0xFF);
		dmawdat(FDC_CS, IRUPT, FDC_DELAY);	/* reset controller */
		curcyl[DRIVE(xp->x_drive)] = -1;
		return(X_ERROR);
	}
	curcyl[DRIVE(xp->x_drive)] = 0;
	return(X_AGAIN);
  default:
	ASSERT(0);
  }
  return(X_OK);
}

/*===========================================================================*
 *				gparam					     *
 *===========================================================================*/
PRIVATE struct fparam *gparam(drive)
int drive;
{
	if (open_count[DRIVE(drive)].formatting) {
		sparam(drive, fparam[DTYPE(drive)][DRIVE(drive)]);
		return(&format);
	}
	return(fparam[DTYPE(drive)][DRIVE(drive)]);
}

/*===========================================================================*
 *				sparam					     *
 *===========================================================================*/
PRIVATE void sparam(drive, f)
int drive; struct fparam *f;
{
	register struct hparam *h;

	h = &hparam[DRIVE(drive)];
	switch (f->density) {
		case DD:
		case DD5:
			format.sector_size = trklen[0];
			break;

		case QD:
			format.sector_size = trklen[1];
			break;

		case HD:
			format.sector_size =
			       trklen[h->size == I5 ?
					2
				:
					3];
			break;
	}
	format.sector_0     = 1;
	format.nr_sectors   = 1;
	format.cylinder_0   = f->cylinder_0;
	format.nr_cylinders = f->nr_cylinders;
	format.nr_sides     = f->nr_sides;
	format.density      = f->density;
	format.stepping     = f->stepping;
}

/*===========================================================================*
 *				do_ioctl				     *
 *===========================================================================*/
PRIVATE int do_ioctl(mp)
register message *mp;		/* pointer to read or write message */
{
  struct fparam ufparam;
  struct hparam uhparam;
  register struct fparam *f;
  register struct hparam *h;
  register phys_bytes uaddress, kaddress;

  if (DRIVE(mp->DEVICE) >= NR_FLOPDRIVES)
	return(ENXIO);
  if (DTYPE(mp->DEVICE) >= NR_TYPES)
	return(ENXIO);
  f = gparam(mp->DEVICE);
  h = &hparam[DRIVE(mp->DEVICE)];

  switch (mp->COUNT) {
	case DIOFMTLOCK:
	case DIOFMTFREE:
		kaddress = uaddress = (phys_bytes) 0;
		break;

	case DIOGETP:
	case DIOSETP:
		if (!(uaddress = numap(mp->PROC_NR, (vir_bytes) mp->ADDRESS,
		      (vir_bytes) sizeof(ufparam))))
			return(EINVAL);
		kaddress = umap(proc_ptr, S, (vir_bytes) &ufparam,
					     (vir_bytes) sizeof(ufparam));
		break;

	case DIOGETHP:
	case DIOSETHP:
		if (!(uaddress = numap(mp->PROC_NR, (vir_bytes) mp->ADDRESS,
		      (vir_bytes) sizeof(uhparam))))
			return(EINVAL);
		kaddress = umap(proc_ptr, S, (vir_bytes) &uhparam,
					     (vir_bytes) sizeof(uhparam));
		break;

	default:
		return(EINVAL);
  }

  switch (mp->COUNT) {
	case DIOFMTLOCK:
		if (open_count[DRIVE(mp->DEVICE)].o_cnt > 1)
			return(EBUSY);
		open_count[DRIVE(mp->DEVICE)].formatting = 1;
		break;

	case DIOFMTFREE:
		open_count[DRIVE(mp->DEVICE)].formatting = 0;
		break;

	case DIOGETP:
		ufparam = *f;
		phys_copy(kaddress, uaddress, (vir_bytes) sizeof(ufparam));
		break;

	case DIOSETP:
		if (open_count[DRIVE(mp->DEVICE)].formatting)
			return(EBUSY);

		phys_copy(uaddress, kaddress, (vir_bytes) sizeof(ufparam));
		if (DTYPE(mp->DEVICE) & 0x0f)	/* only allowed for type 0 */
			return(EINVAL);		/* and type 16             */

		if (ufparam.sector_size !=  128 &&
		    ufparam.sector_size !=  256 &&
		    ufparam.sector_size !=  512 &&
		    ufparam.sector_size != 1024 &&
		    ufparam.sector_size != 2048 &&
		    ufparam.sector_size != 4096)
			return(EINVAL);

		sparam(mp->DEVICE, &ufparam);
		if (ufparam.nr_sectors == 0 ||
		    /*
		     * Not a very good check, but better then none.
		     */
		    ufparam.nr_sectors * ufparam.sector_size > format.sector_size)
			return(EINVAL);

		if (ufparam.cylinder_0 >= h->max_cyl)
			return(EINVAL);

		if ((ufparam.nr_cylinders + ufparam.cylinder_0) > h->max_cyl ||
		    (ufparam.stepping == DSTP &&
		     (ufparam.nr_cylinders + ufparam.cylinder_0) > h->max_cyl/2))
			return(EINVAL);

		if (ufparam.nr_sides > h->sides)
			return(EINVAL);

		if (!(h->dense & (1<<ufparam.density)))
			return(EINVAL);

		if (ufparam.stepping > 1)
			return(EINVAL);

		/*
		 * autocf is not user setable!
		 */
		ufparam.autocf = f->autocf;
		*f = ufparam;
		break;

	case DIOGETHP:
		uhparam = *h;
		phys_copy(kaddress, uaddress, (vir_bytes) sizeof(uhparam));
		break;

	case DIOSETHP:
		if (open_count[DRIVE(mp->DEVICE)].formatting)
			return(EBUSY);

		phys_copy(uaddress, kaddress, (vir_bytes) sizeof(uhparam));
		if (uhparam.max_cyl == 0 ||
		    uhparam.dense == 0 ||
		    uhparam.sides > 1 ||
		    uhparam.slow > 1 ||
		    (uhparam.size != I3 && uhparam.size != I5))
		*h = uhparam;
		break;
  }
  return(OK);
}
#endif
#endif
