/*
 * This file contains a driver for the SCSI Controller on the Atari TT.
 *
 * The driver supports two operations: read a block and write a block.
 * It accepts two messages, one for reading and one for writing,
 * both using message format m2 and with the same parameters:
 *
 *    m_type      DEVICE    PROC_NR     COUNT    POSITION  ADRRESS
 * ----------------------------------------------------------------
 * |  DEV_READ  | device  | proc nr |  bytes  |  offset | buf ptr |
 * |------------+---------+---------+---------+---------+---------|
 * | DEV_WRITE  | device  | proc nr |  bytes  |  offset | buf ptr |
 * |--------------------------------------------------------------|
 * |SCATTERED_IO| device  | proc nr | requests|         | iov ptr |
 * |------------+---------+---------+---------+---------+---------|
 * | DEV_OPEN   | device  | proc nr |         |         |         |
 * |------------+---------+---------+---------+---------+---------|
 * | DEV_CLOSE  | device  | proc nr |         |         |         |
 * ----------------------------------------------------------------
 *
 * The file contains one entry point:
 *
 *	scsi_task: main entry when system is brought up
 *
 * A single TT may have several drives with different scsi id attached to it.
 *
 * Note that the driver is written for the situation where there is only one TT
 * present on the SCSI bus, which acts as initiator. 
 * Using the TT as target, or connecting multiple TT's to use one disk is not
 * supported (although it is technically feasible)
 *
 * TODO:
 * - add initialisation code
 *   use request sense to remove possible unit attention after reset
 *   do an inquiry to determine device type
 * - handle msg in/msg out
 * - fix case where less data were received than requested. DMA controller
 *   then does not interrupt. Accept interrupt of 5380
 */

#include "kernel.h"
#if (MACHINE == ATARI)
#include <minix/callnr.h>
#include <minix/com.h>
#include "proc.h"
#include "staddr.h"
#include "sthdc.h"
#include "stscsi.h"
#include "stscsidma.h"

#if (NR_SCSI_DRIVES == 0)
void scsiint()
{
}

void scsidmaint()
{
}
#else /* NR_SCSI_DRIVES > 0 */

#define TRACE(x)	/* x */

/* Parameters for the disk drive. */
#define SECTOR_SIZE	512	/* physical sector size in bytes */
#define MINOR_SHIFT	4	/* 2 ** MINOR_SHIFT == #minors/drive */
				
/* timing constants */
#define	SCSI_POLL	1000000	/* polling of scsi bits */

/* initiator scsi id bit */
#define	INITIATOR_ID_BIT	0x80	/* initiator has scsi id 7 */

PRIVATE message mess;		/* message buffer for in and out */

_PROTOTYPE( PRIVATE void wait_req, (char *s)				);
_PROTOTYPE( PRIVATE void wait_req_done, (char *s)			);
_PROTOTYPE( PRIVATE void snd, (int c)					);
_PROTOTYPE( PRIVATE int rcv, (void)					);

/*===========================================================================*
 *				scsi_task				     *
 *===========================================================================*/
PUBLIC void scsi_task()
{
  register int r, caller, procno;

  TRACE(printf("scsi: task started\n"));

  TRACE(printf("checking if the scsi bus is free\n"));
    
  if (SCSI_HW->CSB & BSY) {
        /* I've chosen to reset the bus if it is not free
           other alternatives are to return without resetting, 
           or to issue a panic.
           Most likely this condition was caused when the system was
           rebooted when the SCSI bus was busy, so resetting seems to be
           the best action 
        */
    	printf("SCSI bus is busy, resetting it.");
    	SCSI_HW->ICR = ASS_RST;
    	SCSI_HW->ICR = 0;
  }

  /*
   * The main loop of the disk task.
   * It waits for a message, carries it out, and sends a reply.
   */
  while (TRUE) {
	receive(ANY, &mess);
	if (mess.m_source < 0)
		panic("disk task got message from ", mess.m_source);
	TRACE(printf("hd: received %d from %d\n",mess.m_type,mess.m_source));
	caller = mess.m_source;
	procno = mess.PROC_NR;
	mess.DEVICE += NR_ACSI_DRIVES << MINOR_SHIFT;

	switch (mess.m_type) {
	    case DEV_OPEN:	r = wini_open(&mess);	break;
	    case DEV_CLOSE:	r = wini_close(&mess);	break;
	    case DEV_READ:
	    case DEV_WRITE:	r = wini_rdwt(&mess);	break;
#if (SCSI_SCAT_SECTORS == 0)
	    case SCATTERED_IO:	r = do_vrdwt(&mess, wini_rdwt);	break;
#else
	    case SCATTERED_IO:	r = wini_hvrdwt(&mess);	break;
#endif
	    case DEV_IOCTL:	r = wini_ioctl(&mess);	break;
	    default:		r = EINVAL;		break;
	}
	
	/* Finally, prepare and send the reply message. */
	mess.m_type = TASK_REPLY;	
	mess.REP_PROC_NR = procno;
	mess.REP_STATUS = r;	/* # of bytes transferred or error code */
	send(caller, &mess);	/* send reply to caller */
  }
}

/*===========================================================================*
 *				scsidmaint				     *
 *===========================================================================*/
PUBLIC void scsidmaint()
{
  unsigned char control_value;
  char *addr;
  unsigned long endaddr;
  int i;
  
  TRACE(printf("got scsi dma interrupt\n"));
  
  wait_req("no request in dma interrupt");
  SCSI_HW->MR2 = 0;
	
  control_value = SCSIDMA->dma_control;
  if (control_value & BERR) panic("bus error during scsi dma tranfer", NO_NUM);
  
  /* if a dma transfer does not fill a double word the control software must
     copy the data residue register to the destination address;
     dma_addr always contains the address of the last byte which is tranferred
     so its contents can be used to determine which part of the data residue
     register still needs to be written */
     
  endaddr = ((unsigned long)(SCSIDMA->dma_addr[AD_UP]) << 24) +
            ((unsigned long)(SCSIDMA->dma_addr[AD_UP_MID]) << 16) +
            ((unsigned long)(SCSIDMA->dma_addr[AD_LOW_MID]) << 8) +
             (unsigned long)(SCSIDMA->dma_addr[AD_LOW]);
            
  addr = (char *)(endaddr & ~3);
  i = 0;
  
  switch (endaddr & 3) {
  case 3:
  	*addr++ = SCSIDMA->data_residue[i++];
  	/* fall through for next byte */
  case 2:
  	*addr++ = SCSIDMA->data_residue[i++];
  	/* fall through for next byte */
  case 1:
  	*addr++ = SCSIDMA->data_residue[i++];
  	break;
  case 0:
  	break;
  }
  
  interrupt(SCSI);
}

/*===========================================================================*
 *				scsiint					     *
 *===========================================================================*/
PUBLIC void scsiint()
{
  register int temp;
   
  /* this should never occur
     the else is not really needed since panic never returns
     most of the interrupts are disabled because we're not 
     interested in them or because we have other ways to detect
     the event signalled by the interrupt (e.g. for end of dma the dma
     controller also generates an interrupt).
  */
     
  TRACE(printf("got scsi dma interrupt\n"));
  
  temp = SCSI_HW->BSR;
  if ((temp & INT) == 0) panic("spurious scsi interrupt", NO_NUM);
  else
    if (temp & EDMA) panic("end of dma interrupt", NO_NUM);
    else
      if (temp & SPER) panic("scsi parity interrupt", NO_NUM);
      else
        if (temp & PHSM) panic("scsi phase mismatch interrupt", NO_NUM);
        else
           if (temp & MBSY) panic("scsi busy loss interrupt", NO_NUM);
           else 
           {
	     temp = SCSI_HW->CSB & (SEL | IO) ;
	     if (temp == (SEL | IO)) panic("scsi reselection interrupt", NO_NUM);
	     else
	       if (temp & SEL) panic("scsi selection interrupt", NO_NUM);
	       else
	         if (!temp) panic("scsi bus reset interrupt", NO_NUM);
	         else panic("unknown scsi interrupt", NO_NUM);
           }
}

/*===========================================================================*
 *				wait_req				     *
 *===========================================================================*/
 
PRIVATE void wait_req(s)
char *s;
{
    register long i = SCSI_POLL;
    
    /* wait for the REQ line (active low) to be asserted */
    while(!(SCSI_HW->CSB & REQ)) if(!i--) {
/*
    	panic(s, NO_NUM);
*/
printf("wait_req: %s\n", s);
    }
}

/*===========================================================================*
 *				wait_req_done				     *
 *===========================================================================*/
 
PRIVATE void wait_req_done(s)
char *s;
{
    register long i = SCSI_POLL;
    
    /* wait for the REQ line (active low) to be deasserted */
    while(SCSI_HW->CSB & REQ) if(!i--) {
/*
    	panic(s, NO_NUM);
*/
printf("wait_req_done: %s\n", s);
    }
}

/*===========================================================================*
 *				scsi_cmd				     *
 *===========================================================================*/
int scsi_cmd(drive, cmd, cmdlen, address, data_len, rw)
int		drive;
unsigned char  *cmd;
int		cmdlen;
phys_bytes 	address;
phys_bytes data_len;
int		rw;
{
    register long i;
    
    /* initialise some scsi registers */
    
    SCSI_HW->ICR = 0;
    SCSI_HW->TCR = 0;
    
    TRACE(printf("checking if the scsi bus is free\n"));
    
    if (SCSI_HW->CSB & BSY) {
    	return(EBUSY);
    }
    
    /* arbitration phase */
    
    /* since we assume that we are the only initiator there is no arbitration 
       phase. If we are not busy the scsi bus must be free */
    
    /* selection phase */
    
    TRACE(printf("start selection\n"));
    
    SCSI_HW->SER = 0;
    SCSI_HW->ODR = INITIATOR_ID_BIT | (1<<drive);
    SCSI_HW->ICR = ASS_DBUS | ASS_SEL;

    i = SCSI_POLL;
    while(!(SCSI_HW->CSB & BSY)) if(!i--) {
    	printf("No response from scsi device %d.\n", drive);
    	SCSI_HW->ICR = 0;
    	return(EIO);
    }
    SCSI_HW->ICR = 0;
    SCSI_HW->TCR = ASS_CD;
    
    TRACE(printf("Sel"));
    
    /* setup dma */
    
    if (rw == DEV_READ)
    {
        SCSIDMA->dma_control = 0;
    }
    else /* rw == DEV_WRITE */
    {
        SCSIDMA->dma_control = DIR;
    }
    
    SCSIDMA->dma_addr[AD_UP]      = (char) (address >> 24);
    SCSIDMA->dma_addr[AD_UP_MID]  = (char) (address >> 16);
    SCSIDMA->dma_addr[AD_LOW_MID] = (char) (address >>  8);
    SCSIDMA->dma_addr[AD_LOW]     = (char) (address      );
    SCSIDMA->dma_byte_cnt[AD_UP]      = (char) (data_len >> 24);
    SCSIDMA->dma_byte_cnt[AD_UP_MID]  = (char) (data_len >> 16);
    SCSIDMA->dma_byte_cnt[AD_LOW_MID] = (char) (data_len >>  8);
    SCSIDMA->dma_byte_cnt[AD_LOW]     = (char) (data_len      );
    
    while (1) 
    {
	wait_req("No scsi request");
	/* probably return an error if there is no request here */
	switch(SCSI_HW->CSB & CSB_PHASE_MASK)
	{
		case DATA_OUT_PHASE:
			TRACE(printf("DO"));
			SCSI_HW->TCR = 0;
			SCSI_HW->ICR = ASS_DBUS;
			SCSI_HW->MR2 = PINT | PCHK | DMA;
			SCSI_HW->SDS = 0;
			SCSIDMA->dma_control = ENABLE | DIR;
			receive(HARDWARE, &mess);	/* Wait for interrupt. */
			SCSI_HW->ICR = 0;
			break;
		case DATA_IN_PHASE:
			TRACE(printf("DI"));
			SCSI_HW->TCR = 0;
			SCSI_HW->TCR = ASS_IO;
			SCSI_HW->MR2 = PINT | PCHK | DMA;
			SCSI_HW->SDI = 0;
			SCSIDMA->dma_control = ENABLE;
			receive(HARDWARE, &mess);	/* Wait for interrupt. */
		case COMMAND_PHASE:
			TRACE(printf("C"));
			snd(*cmd++);
			break;
		case STATUS_PHASE:
			TRACE(printf("S"));
			i = rcv();
			if (i & 0x1e) panic("scsi status phase didn't return GOOD\n", i);
			break;
		case MESSAGE_OUT_PHASE:
			TRACE(printf("MO"));
			break;
		case MESSAGE_IN_PHASE:
			TRACE(printf("MI"));
			i = rcv();
			if (i) panic("scsi message in phase didn't return COMMAND COMPLETE\n", i);
    			TRACE(printf("dis\n"));
/* Maybe not return here but wait if a new request appears */
			return(OK);
			break;
		default:
			panic("scsi bus is in illegal phase\n", NO_NUM);
	}
    }
}

PRIVATE void snd(c)
int c;
{
    TRACE(printf("s%02x", c));
    
    wait_req("No request for snd");
    SCSI_HW->ODR = c;
    SCSI_HW->ICR = ASS_DBUS;
    SCSI_HW->ICR = ASS_DBUS | ASS_ACK;
    wait_req_done("Request for snd not terminated");
    SCSI_HW->ICR = 0;
}

PRIVATE int rcv()
{
    int c;
    
    wait_req("No request for rcv");
    c = SCSI_HW->CSD;
    
    TRACE(printf("r%02x", c));
    
    /* acknowledge reception */
    SCSI_HW->ICR = ASS_ACK;
    wait_req_done("Request for rcv not terminated");
    SCSI_HW->ICR = 0;

    return(c);
}
#endif
#endif
